@@ -29,7 +29,7 @@ this object structure automatically from a .proto souce file.
 import struct
 
 from extprot.errors import *
-from extprot.stream import StringStream
+from extprot.stream import Stream, StringStream
 
 
 def _issubclass(cls,bases):
@@ -71,7 +71,8 @@ class Type(object):
     def _convert_types(cls,values,types=None):
         """Convert a sequence of values using a type tuple.
 
-        If no type tuple is given, cls._types is used.
+        If no type tuple is given, cls._types is used.  If there aren't
+        enough values for the number of types, we try to use default values.
         """
         values = iter(values)
         if types is None:
@@ -80,7 +81,7 @@ class Type(object):
             try:
                 v = values.next()
             except StopIteration:
-                raise ValueError("too few values to convert")
+                yield t.default()
             else:
                 yield t.convert(v)
         try:
@@ -117,6 +118,11 @@ class Type(object):
         s = StringStream(string)
         return cls.from_stream(s)
 
+    @classmethod
+    def from_file(cls,file):
+        s = Stream(file)
+        return cls.from_stream(s)
+
     def __eq__(self,other):
         return self is other
 
@@ -176,7 +182,9 @@ class Int(Type):
 
     @classmethod 
     def convert(cls,value):
-        return int(value)
+        if not isinstance(value,(int,long)):
+            raise ValueError("not a valid Int: " + repr(value))
+        return value
 
     @classmethod
     def from_stream(cls,stream):
@@ -284,8 +292,6 @@ class Tuple(Type):
         values = []
         for t in cls._types[:nelems]:
             values.append(t.from_stream(stream))
-        for t in cls._types[nelems:]:
-            values.append(t.default())
         for _ in xrange(max(0,nelems - len(cls._types))):
             stream.skip_value()
         return cls._convert_types(values)
@@ -318,9 +324,6 @@ class _TypedList(list):
         items = (self._types.convert(i) for i in sequence)
         super(_TypedList,self).__setslice__(i,j,items)
 
-    def __contains__(self,item):
-        super(_TypedList,self).__contains__(self._type.convert(item))
-
     def append(self,item):
         return super(_TypedList,self).append(self._type.convert(item))
 
@@ -446,7 +449,10 @@ class _UnionMetaclass(type):
                 else:
                     t._index = t_idx
                     t_idx += 1
-        #  Build the prefix lookup table
+                #  Adjust __name__ and __module__ to allow pickling
+                if _issubclass(t,Message):
+                    t.__name__ = cls.__name__+"."+t.__name__
+                    t.__module__ = cls.__module__
         cls._option_from_prefix = {}
         for t in cls._types:
             # TODO: delegate this formatting to the stream somehow?
@@ -494,7 +500,7 @@ class Option(Type):
         """
         if not cls._types:
             if values:
-                raise TypeError("values given to constant Option")
+                raise ValueError("values given to constant Option constructor")
             return cls
         else:
             return Type.__new__(cls)
@@ -509,6 +515,7 @@ class Option(Type):
         self._values[index] = self._types[index].convert(value)
 
     def __eq__(self,other):
+        print "EQ"
         return self._values == other._values
 
     @classmethod
@@ -523,20 +530,24 @@ class Option(Type):
         #  accept instances of any of our base classes, as long as they
         #  have a compatible type signature.
         if isinstance(value,Option):
-            if value.__class__ not in cls.__mro__:
-                raise ValueError("not this Option type")
-            if unify_types(cls._types,value.__class__._types) is None:
-                raise ValueError("not this Option type")
-            return cls(*value._values)
-        if _issubclass(value,Option):
+            if value.__class__ in cls.__mro__:
+                return cls(*value._values)
+            if unify_types(cls._types,value.__class__._types) is not None:
+                return cls(*value._values)
+        elif _issubclass(value,Option):
             if value._types:
                 raise ValueError("no data given to non-constant Option")
-            if value not in cls.__mro__:
-                raise ValueError("not this Option type")
-            if unify_types(cls._types,value._types) is None:
-                raise ValueError("not this Option type")
-            return cls
-        #  Not an Option class or instance, must be a direct value tuple
+            if not cls._types and value  in cls.__mro__:
+                return cls
+        #  Not an Option class or instance, must be a direct value tuple.
+        #  A scalar is converted into a tuple of length 1.
+        if isinstance(value,basestring):
+            value = (value,)
+        else:
+            try:
+                value = iter(value)
+            except TypeError:
+                value = (value,)
         return cls(*value)
 
     @classmethod
@@ -551,9 +562,8 @@ class Option(Type):
         values = []
         for t in cls._types[:nelems]:
             values.append(t.from_stream(stream))
-        for t in cls._types[nelems:]:
+        for _ in xrange(max(0,nelems - len(cls._types))):
             stream.skip_value()
-            values.append(t.default())
         return cls(*cls._convert_types(values))
 
     @classmethod
@@ -589,7 +599,7 @@ class Field(Type):
         self._creation_order = Field._creation_counter
         Field._creation_counter += 1
         self._types = (type,)
-        self.mutable = False
+        self.mutable = mutable
 
     def __get__(self,obj,type=None):
         try:
@@ -601,7 +611,11 @@ class Field(Type):
         if not self.mutable and obj._initialized:
             raise AttributeError("Field '"+self._name+"' is not mutable")
         if value is None:
-            value = self._types[0].default()
+            try:
+                value = self._types[0].default()
+            except UndefinedDefaultError:
+                msg = "value required for field " + self._name
+                raise UndefinedDefaultError(msg)
         obj.__dict__[self._name] = self._types[0].convert(value)
 
     def from_stream(self,stream):
@@ -645,12 +659,13 @@ class Message(Type):
     """Composed message type.
 
     This is the basic unit of data transfer in extprot, and is basically
+    a set of typed key-value pairs.
     """
 
     __metaclass__ = _MessageMetaclass
 
     #  Messages can be part of a Union, giving them a specific index.
-    #  Stand-alone messaged always have index of zero.
+    #  Stand-alone messaged always have an index of zero.
     _index = 0
 
     def __init__(self,*args,**kwds):
@@ -709,7 +724,7 @@ class Message(Type):
 
     @classmethod
     def to_stream(cls,value,stream):
-        values = ((t.to_stream,t.__get__(value)) for t in cls._types)
+        values = ((t.to_stream,value.__dict__[t._name]) for t in cls._types)
         stream.write_Tuple(value._index,values)
 
     def to_string(self):
@@ -717,6 +732,9 @@ class Message(Type):
         self.to_stream(s)
         return s.getstring()
 
+    def to_file(self,file):
+        self.to_stream(Stream(file))
+
     @classmethod
     def from_string(cls,string):
         s = StringStream(string)
@@ -730,6 +748,29 @@ class Message(Type):
                 return False
         return True
 
+    def __reduce__(self):
+        """Pickle Messages by serializing them.
+
+        For this to work, we have to be able to find the Message class
+        after unpickling.  This means that self.__class__.__name__ and
+        self.__class__.__module__ must be set to something useful. In
+        particuler, __name__ on inner classes must be set to the full
+        dotted name of the class.
+        """
+        cls = self.__class__
+        args = (cls.__module__,cls.__name__,self.to_string())
+        return (_unpickle_message,args)
+
+
+def _unpickle_message(module,name,data):
+    """Helper function for unpickled of Message insances."""
+    mname = module.split(".")[-1]
+    cls = __import__(module,fromlist=[mname])
+    for nm in name.split("."):
+        cls = getattr(cls,nm)
+    return cls.from_string(data)
+_unpickle_message.__safe_for_unpickling__ = True
+
 
 class Union(Type):
     """Composed disjoint-union type.
@@ -770,10 +811,11 @@ class Union(Type):
     def convert(cls,value):
         for t in cls._types:
             try:
-                return t.convert(value)
+                v = t.convert(value)
+                return v
             except ValueError, e:
                 pass
-        raise ValueError("could not convert Union type")
+        raise ValueError("could not convert value " + repr(value) + " to Union type " + repr(cls))
 
     @classmethod
     def default(cls):
@@ -795,7 +837,11 @@ class Union(Type):
 
     @classmethod
     def to_stream(cls,value,stream):
-        value.to_stream(value,stream)
+        if isinstance(value,Message):
+            vcls = value.__class__
+        else:
+            vcls = value
+        vcls.to_stream(value,stream)
 
    
 class Unbound(Type):
